'''
Copyright (c) 2022 Algorand Name Service

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
'''

import unittest, time
from algosdk import mnemonic
import json, random, string

import ans_helper as anshelper
from anssdk import constants
from anssdk.resolver import ans_resolver

unittest.TestLoader.sortTestMethodsUsing = None

class TestDotAlgoNameRegistry(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.algod_client = anshelper.SetupClient()
        cls.algod_indexer = anshelper.SetupIndexer()
        cls.app_index = constants.APP_ID
        cls.resolver_obj = ans_resolver(cls.algod_client, cls.algod_indexer)


    def test_name_resolution(self):
        
        account_info = self.resolver_obj.resolve_name('rand')
        self.assertEqual(account_info["owner"], 'RANDGVRRYGVKI3WSDG6OGTZQ7MHDLIN5RYKJBABL46K5RQVHUFV3NY5DUE')

    def test_names_owned_by_address(self):
        
        account_info = self.resolver_obj.get_names_owned_by_address('RANDGVRRYGVKI3WSDG6OGTZQ7MHDLIN5RYKJBABL46K5RQVHUFV3NY5DUE')
        self.assertGreaterEqual(len(account_info), 2)        

# TODO: See where tearDown goes, class or outside
def tearDownClass(self) -> None:
    # TODO: clear all variables?
    return super().tearDown()

if __name__ == '__main__':
    unittest.main()



'''
Copyright (c) 2022 Algorand Name Service

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
'''

from pyteal import *
from anssdk import constants

def ValidateRecord(name, reg_app_id, reg_escrow_acct):

    DOT_ALGO_APP_ID = reg_app_id
    DOT_ALGO_ESCROW_ADDRESS = reg_escrow_acct
    
    i = ScratchVar(TealType.uint64)

    is_valid_txn = Seq([

        Assert(Len(Bytes(name)) <= Int(64)),
        For(i.store(Int(0)), i.load() < Len(Bytes(name)), i.store(i.load() + Int(1))).Do(
            Assert(
                Or(
                    And(
                        GetByte(Bytes(name), i.load()) >= Int(constants.ASCII_LOWER_CASE_A),
                        GetByte(Bytes(name), i.load()) <= Int(constants.ASCII_LOWER_CASE_Z)
                    ),
                    And(
                        GetByte(Bytes(name), i.load()) >= Int(constants.ASCII_DIGIT_0),
                        GetByte(Bytes(name), i.load()) <= Int(constants.ASCII_DIGIT_9)
                    )
                )
            )
        ),
        
        Assert(
            Or(
                Global.group_size() == Int(2),
                Global.group_size() == Int(4)
            )
        ),
        
        Assert(Gtxn[0].sender() == Gtxn[1].sender()),
        Assert(Gtxn[0].receiver() == Addr(DOT_ALGO_ESCROW_ADDRESS)),

        If(Global.group_size() == Int(2))
        .Then(
            Assert(
                And(

                    Gtxn[1].application_id() == Int(DOT_ALGO_APP_ID),
                    Gtxn[1].application_args[0] == Bytes("register_name"),
                    Gtxn[1].application_args[1] == Bytes(name)
                )
            )
        ).ElseIf(Global.group_size() == Int(4))
        .Then(
            Assert(
                And(

                    Gtxn[1].receiver() == Gtxn[2].sender(),
                    Gtxn[2].application_id() == Int(DOT_ALGO_APP_ID),
                    Gtxn[2].on_completion() == OnComplete.OptIn,
                    Gtxn[3].application_id() == Int(DOT_ALGO_APP_ID),
                    Gtxn[3].sender() == Gtxn[0].sender(),
                    Gtxn[3].application_args[0] == Bytes("register_name"),
                    Gtxn[3].application_args[1] == Bytes(name)
                )
            )
        ).Else(
            Return(Int(0))
        ),

        Int(1)     
    ])

    payment_for_3 = Seq([
        Assert(Gtxn[0].amount() >= Int(constants.COST_FOR_3)),
        Assert(is_valid_txn),
        Int(1)
    ])

    payment_for_4 = Seq([
        Assert(Gtxn[0].amount() >= Int(constants.COST_FOR_4)),
        Assert(is_valid_txn),
        Int(1)
    ])

    payment_for_5 = Seq([
        Assert(Gtxn[0].amount() >= Int(constants.COST_FOR_5)),
        Assert(is_valid_txn),
        Int(1)
    ])

    program = Cond(
        [Len(Bytes(name)) == Int(3), payment_for_3],
        [Len(Bytes(name)) == Int(4), payment_for_4],
        [Len(Bytes(name)) >= Int(5), payment_for_5],
    )

    return program

'''
with open('ans-dot-algo-name-record.teal', 'w') as f:
    compiled = compileTeal(ValidateRecord("ans"), Mode.Signature, version=4)
    f.write(compiled)
'''



'''
Copyright (c) 2022 Algorand Name Service

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
'''

from algosdk import mnemonic, account, encoding
from algosdk.future import transaction
from algosdk.future.transaction import LogicSig, LogicSigTransaction, LogicSigAccount
from algosdk.v2client import algod, indexer
from algosdk import logic
from pyteal import *

import sys
from anssdk import constants
sys.path.append('../')

from anssdk.dot_algo_name_record import ValidateRecord

import base64
import datetime,time

def SetupClient():
    api_key = ""
    with open('api_key.txt', 'r') as f:
        api_key = f.readlines()[0]

    # Purestake conn
    algod_address = "https://mainnet-algorand.api.purestake.io/ps2"
    headers = {
    "X-API-Key": api_key
    }
    
    algod_client=algod.AlgodClient(api_key, algod_address, headers=headers)
    return algod_client

def SetupIndexer():
    api_key = ""
    with open('api_key.txt', 'r') as f:
        api_key = f.readlines()[0]

    algod_address = "https://mainnet-algorand.api.purestake.io/idx2"
    headers = {
        'X-API-key' : api_key,
    }
    algod_indexer=indexer.IndexerClient("", algod_address, headers)
    
    return algod_indexer



from audioop import add
from algosdk import encoding
from pyteal import compileTeal, Mode
from algosdk.future.transaction import LogicSig
from algosdk import logic
import base64
import time
from pyteal import *
from anssdk import constants, dot_algo_name_record

class ans_resolver:
    
    def __init__(self, client, indexer=None):
        self.algod_client = client
        if(indexer is not None):
            self.algod_indexer = indexer

    def compile_program(self, algod_client, source_code) :
        compile_response = algod_client.compile(source_code.decode('utf-8'))
        return base64.b64decode(compile_response['result'])            

    def prep_name_record_logic_sig(self, name, reg_app_id):
        reg_escrow_acct = logic.get_application_address(reg_app_id)
        logic_sig_teal = compileTeal(dot_algo_name_record.ValidateRecord(name,reg_app_id,reg_escrow_acct), Mode.Signature, version=4)
        validate_name_record_program = self.compile_program(self.algod_client, str.encode(logic_sig_teal))
        lsig = LogicSig(validate_name_record_program)
        return lsig

    def resolve_name(self, name):
        
        name = name.split('.algo')[0]
        reg_app_id = constants.APP_ID
        
        try:
            for apps_local_data in self.algod_client.account_info(address=self.prep_name_record_logic_sig(name, reg_app_id).address())['apps-local-state']:
                owner = None
                expiry = None
                if(apps_local_data['id']==reg_app_id):
                    for key_value in apps_local_data['key-value']:
                        if(base64.b64decode(key_value['key']).decode()=="expiry"):
                            expiry = key_value['value']['uint']
                        elif(base64.b64decode(key_value['key']).decode()=="owner"):
                            owner = encoding.encode_address(base64.b64decode(key_value['value']['bytes']))
                if(owner!=None and expiry!=None and expiry>int(time.time())):
                    return ({
                        'found': True,
                        'owner': owner,
                        'expiry': expiry
                    })
                else:
                    return None

        except Exception as err:
            return ({
                'found': False,
                'error': err
            })         

    def get_names_owned_by_address(self,address):
        is_valid_address = encoding.is_valid_address(address)
        indexer = None
        try:
            if(self.algod_indexer is not None):
                indexer = self.algod_indexer
        except Exception as err:
            return({
                'err': 'Indexer is not instantiated'
            })
            return

        if(is_valid_address is False):
            return(
                {
                    'err':'Invalid Algorand address'
                }
            )
        else:
            
            next_token = ''
            txn_length = 1
            txns = []
            
            while(txn_length > 0):
                
                account_info = indexer.search_transactions_by_address(address=address, limit=10000, next_page=next_token, start_time="2022-02-25")
                
                if(len(account_info["transactions"]) > 0):
                    txn_length = len(account_info["transactions"])
                    txns+=account_info["transactions"]
                    if(account_info["next-token"] is not None):
                        next_token = account_info["next-token"]
                    else:
                        break
                else:
                    break
                
            names = []
            try:
                for txn in txns:
                    
                    if(txn["tx-type"] == "appl"):
                        
                        if(txn["application-transaction"]["application-id"] == constants.APP_ID):
                            args = txn["application-transaction"]["application-args"]
                            method = base64.b64decode(args[0]).decode('utf-8')
                            if(len(args) >= 2 or method == 'accept_transfer'):
                                
                                if(method == 'register_name'):
                                    arg_1 = base64.b64decode(args[1]).decode('utf-8')
                                    if(arg_1 not in names):
                                        names.append(arg_1+'.algo')
                                elif(method == 'accept_transfer'):
                                    lsig_account = txn["application-transaction"]["accounts"][0]
                                    lsig_account_info = indexer.account_info(lsig_account)
                                    lsig_account_info = lsig_account_info["account"]["apps-local-state"]
                                    
                                    for app in lsig_account_info:
                                        if(app["id"] == constants.APP_ID):
                                            kv_pairs = app["key-value"]
                                            
                                            for kv_pair in kv_pairs:
                                                
                                                key = base64.b64decode(kv_pair["key"]).decode('utf-8')
                                                value = base64.b64decode(kv_pair["value"]["bytes"])
                                                #print(key)
                                                if(key == 'name'):
                                                    
                                                    if(value not in names):
                                                        names.append(value.decode('utf-8')+'.algo')
                                                        break



            except Exception as err:
                print('Error: ',err)
                return []

            return names            
        
    








from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.0.4'
DESCRIPTION = 'Python SDK to resolve ANS .algo names'
LONG_DESCRIPTION = 'This Python SDK helps in resolving ANS .algo names to wallet addresses and view names owned by a specific Algorand Wallet address'

# Setting up
setup(
    name="anssdk",
    version=VERSION,
    author="Algorand Name Service",
    author_email="contact@algonameservice.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=[],
    keywords=['python', 'algorand', 'algorand name service', 'py-sdk', 'name service', '.algo'],
    
)


COST_FOR_3 = 150000000
COST_FOR_4 = 50000000
COST_FOR_5 = 5000000
COST_FOR_TRANSFER = 2000000
COST_FOR_RENEWAL = 5000000
RENEWAL_TIME = 86400*365
ASCII_LOWER_CASE_A = 97
ASCII_LOWER_CASE_Z = 122
ASCII_DIGIT_0 = 48
ASCII_DIGIT_9 = 57
APP_ID = 628095415
RELEASE_TIME = "2022-02-25T10:00:00"